{
*###########################################################################
* This program is used to test the basic functions of the SD file system.
* It implements simple versions of the cat, rm, ls, echo, cd, pwd, mkdir and
* rmdir commands plus the <, > and >> file redirection operators.
* The program starts up the file driver and then prompts for a command.
*
* Written by Dave Hein
* Copyright (c) 2011 Parallax, Inc.
* MIT Licensed
*###########################################################################
}

CON
  _clkmode = xtal1+pll16x
  _clkfreq = 80_000_000

OBJ
  c : "clibsd"
  
VAR
  long stdinfile
  long stdoutfile

PUB start
  'waitcnt(3*clkfreq+cnt)
  c.start
  c.printf0(string("Calling main\n"))
  main

{ Print help information }
PUB Help
  c.printf0(string("Commands are help, cat, rm, ls, ll, echo, cd, pwd, mkdir and rmdir\n"))

PUB Cd(argc, argv)
  if (argc < 2)
    return
  if (c.chdir(long[argv][1]))
    c.perror(long[argv][1])

PUB Pwd(argc, argv) | buffer[16]
  c.getcwd(@buffer, 64)
  c.fprintf1(stdoutfile, string("%s\n"), @buffer)

PUB Mkdir(argc, argv) | i
  i := 1
  repeat while (i < argc)
    if (c.mkdir(long[argv][i], -1))
      c.perror(long[argv][i])
    i++

PUB Rmdir(argc, argv) | i
  i := 1
  repeat while (i < argc)
    if (c.rmdir(long[argv][i]))
      c.perror(long[argv][i])
    i++

{ This routine implements the file cat function }
PUB Cat(argc, argv) | i, num, infile, buffer[10]
  i := 0
  repeat while (i < argc)
    if (i == 0)
      if (argc == 1 or stdinfile <> c.getstdin)
        infile := stdinfile
      else
        i++
        next
    else
      infile := c.fopen(long[argv][i], string("r"))
      if (infile == 0)
        c.perror(long[argv][i])
        i++
        next
    if (infile == c.getstdin)
      repeat while (c.gets(@buffer))
        if (byte[@buffer][0] == 4)
          quit
        c.fprintf1(stdoutfile, string("%s\n"), @buffer)
    else
      repeat while ((num := c.fread(@buffer, 1, 40, infile)) > 0)
        'c.printf1(string("num = %d\n"), num)
        c.fwrite(@buffer, 1, num, stdoutfile)
    if (i)
      c.fclose(infile)
    i++
  c.fflush(c.getstdout)

{ This routine deletes the files specified by the command line arguments }
PUB Remove_0(argc, argv) | i
  i := 1
  repeat while (i < argc)
    if (c.remove(long[argv][i]))
      c.perror(long[argv][i])
    i++

{ This routine echos the command line arguments }
PUB Echo(argc, argv) | i
  i := 1
  repeat while (i < argc)
    if (i <> argc - 1)
      c.fprintf1(stdoutfile, string("%s "), long[argv][i])
    else
      c.fprintf1(stdoutfile, string("%s\n"), long[argv][i])
    i++

{ This routine lists the root directory or any subdirectories specified
   in the command line arguments.  If the "-l" option is specified, it
   will print the file attributes and size.  Otherwise, it will just
   print the file names.  }
PUB List(argc, argv) | i, j, ptr, fname[4], count, filesize, longflag, path, drwx[2], column, prevlen, dirp, entry
  count := 0
  longflag := 0
  ' Check flags
  j := 1
  repeat while (j < argc)
    if (byte[long[argv][j]][0] == "-")
      if (not c.strcmp(long[argv][j], string("-l")))
        longflag := 1
      else
        c.printf1(string("Unknown option \042%s\042\n"), long[argv][j])
    else
      count++
    j++
  ' List directories
  j := 1
  repeat while (j < argc or count == 0)
    if (count == 0)
      count--
      path := string(".")
    elseif (byte[long[argv][j]][0] == "-")
      j++
      next
    else
      path := long[argv][j]
    if (count => 2)
      c.fprintf1(stdoutfile, string("\n%s:\n"), path)
    dirp := c.opendir(path)
    if (not dirp)
      c.perror(path)
      j++
      next
    column := 0
    prevlen := 14
    repeat while (entry := c.readdir(dirp))
      c.strcpy(@fname, @long[entry][2])
      if byte[@fname] == "."
        next
      filesize := long[entry][1]
      c.strcpy(@drwx, string("----"))
      if long[entry] & 8
        byte[@drwx][0] := "d"
      if long[entry] & 4
        byte[@drwx][1] := "r"
      if long[entry] & 2
        byte[@drwx][2] := "w"
      if long[entry] & 1
        byte[@drwx][3] := "x"
      if (longflag)
        c.fprintf3(stdoutfile, string("%s %8d %s\n"), @drwx, filesize, @fname)
      elseif (++column == 5)
        i := prevlen
        repeat while (i < 14)
          c.fprintf0(stdoutfile, string(" "))
          i++
        c.fprintf1(stdoutfile, string("%s\n"), @fname)
        column := 0
        prevlen := 14
      else
        i := prevlen
        repeat while (i < 14)
          c.fprintf0(stdoutfile, string(" "))
          i++
        prevlen := strsize(@fname)
        c.fprintf1(stdoutfile, string("%s"), @fname)
    c.closedir(dirp)
    if (not longflag and column)
      c.fprintf0(stdoutfile, string("\n"))
    j++

{ This routine returns a pointer to the first character that doesn't
   match val. }
PUB SkipChar(ptr, val)
  repeat while (byte[ptr])
    if (byte[ptr] <> val)
      quit
    ptr++
  return ptr

{ This routine returns a pointer to the first character that matches val. }
PUB FindChar(ptr, val)
  repeat while (byte[ptr])
    if (byte[ptr] == val)
      quit
    ptr++
  return ptr

{ This routine extracts tokens from a string that are separated by one or
   more spaces.  It returns the number of tokens found. }
PUB tokenize(ptr, tokens) | num
  num := 0
  repeat while (byte[ptr])
    ptr := SkipChar(ptr, " ")
    if (byte[ptr] == 0)
      quit
    if (byte[ptr][0] == ">")
      ptr++
      if (byte[ptr][0] == ">")
        long[tokens][num++] := string(">>")
        ptr++
      else
        long[tokens][num++] := string(">")
      next
    if (byte[ptr][0] == "<")
      ptr++
      long[tokens][num++] := string("<")
      next
    long[tokens][num++] := ptr
    ptr := FindChar(ptr, " ")
    if (byte[ptr])
      byte[ptr++] := 0
  return num

{ This routine searches the list of tokens for the redirection operators
   and opens the files for input, output or append depending on the 
   operator. }
PUB CheckRedirection(tokens, num) | i, j
  i := 0
  repeat while (i < num - 1)
    if (not c.strcmp(long[tokens][i], string(">")))
      stdoutfile := c.fopen(long[tokens][i + 1], string("w"))
      if (not stdoutfile)
        c.perror(long[tokens][i + 1])
        stdoutfile := c.getstdout
        return 0
    elseif (not c.strcmp(long[tokens][i], string(">>")))
      stdoutfile := c.fopen(long[tokens][i + 1], string("a"))
      if (not stdoutfile)
        c.perror(long[tokens][i + 1])
        stdoutfile := c.getstdout
        return 0
    elseif (not c.strcmp(long[tokens][i], string("<")))
      stdinfile := c.fopen(long[tokens][i + 1], string("r"))
      if (not stdinfile)
        c.perror(long[tokens][i + 1])
        stdinfile := c.getstdin
        return 0
    else
      i++
      next
    j := i + 2
    repeat while (j < num)
      long[tokens][j - 2] := long[tokens][j]
      j++
    i--
    num -= 2
    i++
  return num

{ This routine closes files that were open for redirection }
PUB CloseRedirection
  if (stdinfile <> c.getstdin)
    c.fclose(stdinfile)
    stdinfile := c.getstdin
  if (stdoutfile <> c.getstdout)
    c.fclose(stdoutfile)
    stdoutfile := c.getstdout

PUB mount | mountErr
  c.printf0(string("Load and mount SD: "))
  ' Important: This code assumes you're using a C3 card.
  ' If you're using different hardware, make sure you
  ' change the following initialization to match your card!
  'mountErr := c.mount_explicit(10, 11, 9, 25)
  mountErr := c.mount_explicit(28, 29, $8000, $10000)
  ifnot (mountErr)
    c.printf1(string("Mount error: %d\n"), mountErr)
    c.exit(1)
  c.printf0(string("done.\n\n"))

{ The program starts the file system.  It then loops reading commands
   and calling the appropriate routine to process it. }
PUB main | num, tokens[20], buffer[20]
  stdinfile := c.getstdin
  stdoutfile := c.getstdout
  ' Wait for the serial terminal to start
  'waitcnt(CNT + CLKFREQ)
  mount
  Help
  repeat
    c.printf0(string("\n> "))
    c.fflush(c.getstdout)
    c.gets(@buffer)
    num := tokenize(@buffer, @tokens)
    num := CheckRedirection(@tokens, num)
    if (num == 0)
      next
    if (not c.strcmp(tokens[0], string("help")))
      Help
    elseif (not c.strcmp(tokens[0], string("cat")))
      Cat(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("ls")))
      List(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("ll")))
      tokens[num++] := string("-l")
      List(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("rm")))
      Remove_0(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("echo")))
      Echo(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("cd")))
      Cd(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("pwd")))
      Pwd(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("mkdir")))
      Mkdir(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("rmdir")))
      Rmdir(num, @tokens)
    else
      c.printf0(string("Invalid command\n"))
      Help
    CloseRedirection
  return 0

{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}
